@@ -11,7 +11,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
+import copy
 import fnmatch
 import gc
 import re
@@ -105,7 +105,6 @@ class PPOTrainerTester(unittest.TestCase):
         set_seed(42)
         cls._token = CI_HUB_USER_TOKEN
         cls._api = HfApi(endpoint=CI_HUB_ENDPOINT)
-        cls._api.set_access_token(CI_HUB_USER_TOKEN)
         HfFolder.save_token(CI_HUB_USER_TOKEN)
 
         # model_id
@@ -680,6 +679,47 @@ class PPOTrainerTester(unittest.TestCase):
                 f"Parameter {name} has a gradient larger than max_grad_norm",
             )
 
+    def test_ppo_trainer_kl_penalty(self):
+        dummy_dataset = self._init_dummy_dataset()
+
+        log_probs = torch.Tensor([[0.5, 0.2, 0.1], [0.6, 0.2, 0.1]])
+        ref_log_probs = torch.Tensor([[0.4, 0.3, 0.0], [0.7, 0.1, 0.3]])
+
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=self.gpt2_model,
+            ref_model=None,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        expected_output = torch.Tensor([[0.1000, -0.1000, 0.1000], [-0.1000, 0.1000, -0.2000]])
+        self.assertTrue(torch.allclose(ppo_trainer._kl_penalty(log_probs, ref_log_probs), expected_output))
+
+        self.ppo_config.kl_penalty = "abs"
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=self.gpt2_model,
+            ref_model=None,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        expected_output = torch.Tensor([[0.1000, 0.1000, 0.1000], [0.1000, 0.1000, 0.2000]])
+        self.assertTrue(torch.allclose(ppo_trainer._kl_penalty(log_probs, ref_log_probs), expected_output))
+
+        self.ppo_config.kl_penalty = "mse"
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=self.gpt2_model,
+            ref_model=None,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        expected_output = torch.Tensor([[0.0050, 0.0050, 0.0050], [0.0050, 0.0050, 0.0200]])
+        self.assertTrue(torch.allclose(ppo_trainer._kl_penalty(log_probs, ref_log_probs), expected_output))
+
     @require_peft
     @mark.peft_test
     def test_peft_model_ppo_trainer(self):
@@ -740,6 +780,98 @@ class PPOTrainerTester(unittest.TestCase):
             else:
                 self.assertTrue(param.grad is None, f"Parameter {name} has a gradient")
 
+    @require_peft
+    @mark.peft_test
+    def test_peft_model_ppo_adapter_rm_trainer(self):
+        from peft import LoraConfig, get_peft_model
+        from transformers import AutoModelForCausalLM, AutoModelForSequenceClassification
+
+        dummy_inputs = torch.LongTensor([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]])
+        rm_lora_config = LoraConfig(
+            r=16,
+            lora_alpha=32,
+            lora_dropout=0.05,
+            bias="none",
+            task_type="SEQ_CLS",
+        )
+
+        reward_model = AutoModelForSequenceClassification.from_pretrained(self.model_id)
+        reward_model = get_peft_model(reward_model, rm_lora_config)
+        dummy_optim = torch.optim.Adam(filter(lambda p: p.requires_grad, reward_model.parameters()), lr=1e-3)
+
+        previous_rm_logits = reward_model(dummy_inputs).logits
+        loss = previous_rm_logits.mean()
+        loss.backward()
+
+        dummy_optim.step()
+        reward_model.eval()
+
+        original_rm_logits = reward_model(dummy_inputs).logits
+
+        with tempfile.TemporaryDirectory() as tmpdirname:
+            reward_model.save_pretrained(tmpdirname)
+
+            lora_config = LoraConfig(
+                r=16,
+                lora_alpha=32,
+                lora_dropout=0.05,
+                bias="none",
+                task_type="CAUSAL_LM",
+            )
+            gpt2_model = AutoModelForCausalLM.from_pretrained(self.model_id)
+
+            # this line is very important
+            def make_inputs_require_grad(module, input, output):
+                output.requires_grad_(True)
+
+            gpt2_model.get_input_embeddings().register_forward_hook(make_inputs_require_grad)
+
+            peft_model = get_peft_model(gpt2_model, lora_config)
+            model = AutoModelForCausalLMWithValueHead.from_pretrained(
+                peft_model,
+                reward_adapter=tmpdirname,
+            )
+
+            dummy_dataset = self._init_dummy_dataset()
+            self.ppo_config.batch_size = 2
+            self.ppo_config.mini_batch_size = 1
+
+            ppo_trainer = PPOTrainer(
+                config=self.ppo_config,
+                model=model,
+                ref_model=None,
+                tokenizer=self.gpt2_tokenizer,
+                dataset=dummy_dataset,
+            )
+
+            self.assertTrue(ppo_trainer.ref_model is None)
+
+            dummy_dataloader = ppo_trainer.dataloader
+
+            # train model with ppo
+            for query_tensor, response_tensor in dummy_dataloader:
+                # define a reward for response
+                # (this could be any reward such as human feedback or output from another model)
+                reward = [torch.tensor(1.0), torch.tensor(0.0)]
+                # train model by running a step twice
+                _ = ppo_trainer.step([q for q in query_tensor], [r for r in response_tensor], reward)
+
+                ppo_trainer.model.train()
+                ppo_trainer.model.gradient_checkpointing_enable()
+                _ = ppo_trainer.step([q for q in query_tensor], [r for r in response_tensor], reward)
+                break
+
+            new_logits = ppo_trainer.model.compute_reward_score(dummy_inputs)
+            self.assertTrue(not torch.allclose(previous_rm_logits, new_logits[:, -1, :]))
+            self.assertTrue(torch.allclose(original_rm_logits, new_logits[:, -1, :]))
+
+            # check gradients
+            for name, param in model.named_parameters():
+                if ("lora" in name or "v_head" in name) and ("reward" not in name):
+                    self.assertTrue(param.grad is not None, f"Parameter {name} has a no gradient")
+                else:
+                    self.assertTrue(param.grad is None, f"Parameter {name} has a gradient")
+
     @unittest.skip("Fix by either patching `whomai()` to work in the staging endpoint or use a dummy prod user.")
     def test_push_to_hub(self):
         REPO_NAME = "test-ppo-trainer"
@@ -845,3 +977,124 @@ class PPOTrainerTester(unittest.TestCase):
                 self.assertTrue(param.grad is not None, f"Parameter {name} has a no gradient")
             else:
                 self.assertTrue(param.grad is None, f"Parameter {name} has a gradient")
+
+    def test_generation(self):
+        dummy_dataset = self._init_dummy_dataset()
+
+        model = AutoModelForCausalLMWithValueHead.from_pretrained("gpt2")
+        tokenizer = AutoTokenizer.from_pretrained("gpt2")
+
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=model,
+            ref_model=None,
+            tokenizer=tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        input_texts = ["this is a test", "this is another, longer test"]
+
+        generation_kwargs = {"do_sample": False, "max_new_tokens": 4, "pad_token_id": tokenizer.eos_token_id}
+
+        tokenizer.pad_token = tokenizer.eos_token
+
+        model_inputs = [tokenizer(txt, return_tensors="pt").input_ids.squeeze() for txt in input_texts]
+
+        generations_batched = ppo_trainer.generate(model_inputs, batch_size=2, **generation_kwargs)
+        generations_batched = tokenizer.batch_decode(generations_batched)
+
+        generations_single = [ppo_trainer.generate(inputs, **generation_kwargs).squeeze() for inputs in model_inputs]
+        generations_single = tokenizer.batch_decode(generations_single)
+
+        self.assertEqual(generations_single, generations_batched)
+
+    def test_grad_accumulation(self):
+        dummy_dataset = self._init_dummy_dataset()
+
+        torch.manual_seed(0)
+        gpt2_model = AutoModelForCausalLMWithValueHead.from_pretrained(self.model_id, summary_dropout_prob=0.0)
+        gpt2_model_clone = copy.deepcopy(gpt2_model)
+
+        self.ppo_config.mini_batch_size = 2
+        self.ppo_config.ppo_epochs = 1
+
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=gpt2_model,
+            ref_model=None,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        dummy_dataloader = ppo_trainer.dataloader
+
+        # train model with ppo
+        for query_tensor, response_tensor in dummy_dataloader:
+            # define a reward for response
+            # (this could be any reward such as human feedback or output from another model)
+            reward = [torch.tensor(1.0), torch.tensor(1.0)]
+            # train model by running a step twice
+            _ = ppo_trainer.step([q for q in query_tensor], [r for r in response_tensor], reward)
+            break
+
+        model_grad = gpt2_model.v_head.summary.weight.grad.clone()
+
+        self.ppo_config.gradient_accumulation_steps = 2
+        self.ppo_config.mini_batch_size = 1
+
+        ppo_trainer = PPOTrainer(
+            config=self.ppo_config,
+            model=gpt2_model_clone,
+            ref_model=None,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        dummy_dataloader = ppo_trainer.dataloader
+
+        # train model with ppo
+        for query_tensor, response_tensor in dummy_dataloader:
+            # define a reward for response
+            # (this could be any reward such as human feedback or output from another model)
+            reward = [torch.tensor(1.0), torch.tensor(1.0)]
+            # train model by running a step twice
+            _ = ppo_trainer.step([q for q in query_tensor], [r for r in response_tensor], reward)
+            break
+
+        model_grad_acc = gpt2_model_clone.v_head.summary.weight.grad.clone()
+        self.assertTrue(torch.allclose(model_grad_acc, model_grad, rtol=1e-3, atol=1e-3))
+
+    @unittest.skip("Fix by either patching `whomai()` to work in the staging endpoint or use a dummy prod user.")
+    def test_push_to_hub_if_best_reward(self):
+        REPO_NAME = "test-ppo-trainer"
+        repo_id = f"{CI_HUB_USER}/{REPO_NAME}"
+
+        dummy_dataset = self._init_dummy_dataset()
+
+        push_to_hub_if_best_kwargs = {"repo_id": repo_id}
+
+        ppo_config = PPOConfig(
+            batch_size=2,
+            mini_batch_size=1,
+            log_with=None,
+            push_to_hub_if_best_kwargs=push_to_hub_if_best_kwargs,
+            compare_steps=1,
+        )
+
+        ppo_trainer = PPOTrainer(
+            config=ppo_config,
+            model=self.gpt2_model,
+            ref_model=self.gpt2_model_ref,
+            tokenizer=self.gpt2_tokenizer,
+            dataset=dummy_dataset,
+        )
+
+        dummy_dataloader = ppo_trainer.dataloader
+        # train model with ppo
+        for query_tensor, response_tensor in dummy_dataloader:
+            # define a reward for response
+            # (this could be any reward such as human feedback or output from another model)
+            reward = [torch.tensor(1.0), torch.tensor(0.0)]
+            # train model
+            _ = ppo_trainer.step([q for q in query_tensor], [r for r in response_tensor], reward)
+            break
